@@ -75,6 +75,7 @@ impl fmt::Display for FixedBytesWriterError {
 }
 
 impl Error for FixedBytesWriterError {
+    #[inline]
     fn custom<T>(message: T) -> Self
     where
         T: 'static + Send + Sync + fmt::Display + fmt::Debug,
@@ -82,7 +83,8 @@ impl Error for FixedBytesWriterError {
         Self(FixedBytesWriterErrorRepr::collect(message))
     }
 
-    fn collect_from_display<T>(message: T) -> Self
+    #[inline]
+    fn message<T>(message: T) -> Self
     where
         T: fmt::Display,
     {
@@ -99,9 +101,10 @@ where
 {
     type Error = E;
 
+    #[inline]
     fn write_bytes(&mut self, bytes: &[u8]) -> Result<(), Self::Error> {
         if bytes.len() > N.saturating_sub(self.init) {
-            return Err(E::collect_from_display(BufferOverflow {
+            return Err(E::message(BufferOverflow {
                 at: self.init,
                 additional: bytes.len(),
                 capacity: N,
